from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Union

from .cache import SqliteCache

if TYPE_CHECKING:
    from esi_requests.models import ESIResponse


@dataclass
class _ETagCacheEntry:
    etag: str
    response: "ESIResponse"

class ETagCache:
    """Caches ETags and their associated ESIResponses. 
    Cache keys are URLs, and values are ``_ETagCacheEntry`` objects.
    """
    def __init__(self) -> None:
        self.cache = SqliteCache("request_cache", table="etag")
    
    def get(self, url: str, default: Any = None) -> "_ETagCacheEntry":
        """Gets the ETag response for a given URL."""
        return self.cache.get(url, default)

    def set(self, url: str, etag: str, response: "ESIResponse", expires: Union[str, int]):
        """Sets the ETag response for a given URL."""
        entry = _ETagCacheEntry(etag, response)
        self.cache.set(url, entry, expires)



import base64
import hashlib
import secrets
from typing import Any

from .shared_flow import (
    generate_auth_url,
    send_token_request,
    handle_sso_token_response,
)


def esi_oauth_local(**kwd) -> Any:
    """A local implementation of the OAuth 2.0 native flow.
    Source: https://github.com/esi/esi-docs/blob/master/examples/python/sso/esi_oauth_native.py"""

    callbackURL = kwd.get("callbackURL", "https://localhost/callback/")

    if "clientID" not in kwd.keys() or "scope" not in kwd.keys():
        raise ValueError("ESI Oauth requied field not given.")

    client_id = kwd["clientID"]

    scope = kwd["scope"]

    # Generate the PKCE code challenge
    random = base64.urlsafe_b64encode(secrets.token_bytes(32))
    m = hashlib.sha256()
    m.update(random)
    d = m.digest()
    code_challenge = base64.urlsafe_b64encode(d).decode().replace("=", "")

    generate_auth_url(
        client_id,
        code_challenge,
        scope=scope,
        callbackURL=callbackURL,
    )

    print("\nAuthorizing following scopes:")
    for _scp in scope.split(" "):
        print("\t", _scp)
    print(
        "Authorization url copied to clipboard... Log into your character to grant scopes above."
    )
    auth_code = input("Copy the entire URL and paste it here: ")
    start_index = auth_code.find("?code=") + len("?code=")
    end_index = auth_code.find("&state")  # this identifier might be volatile
    auth_code = auth_code[start_index:end_index]

    code_verifier = random

    form_values = {
        "grant_type": "authorization_code",
        "client_id": client_id,
        "code": auth_code,
        "code_verifier": code_verifier,
    }

    res = send_token_request(form_values)

    return handle_sso_token_response(res)



import requests
import time

from esi_requests.log import getLogger
logger = getLogger(__name__)


def refresh_token(refresh_token: str, clientId: str):
    """Refresh a token with clientId.
    Source: https://github.com/esi/esi-docs/tree/master/examples/python/sso"""

    base_auth_url = "https://login.eveonline.com/v2/oauth/token"

    form_value = {
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
        "client_id": clientId,
    }
    headers = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Host": "login.eveonline.com",
    }

    res = requests.post(base_auth_url, data=form_value, headers=headers)

    data = res.json()
    data["retrieve_time"] = int(time.time())

    logger.debug("Token refresh successful: %s", clientId)
    return data



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .models import ESIRequest


class InvalidRequestError(Exception):
    """Incorrect request parameters being blocked."""

    def __init__(self, reson: str, value):
        self.reason = reson
        self.value = value
        self.msg = "Request blocked: invalid request parameter {reason}: {value}"

    def __str__(self):
        return self.msg.format(reason=self.reason, value=self.value)


class ESIResponseError(Exception):
    """Connection error during reading response, catch from aiohttp.ClientResponseError.

    request_info: instance of ESIRequest
    """
    def __init__(self, status: int, request_info: "ESIRequest", message: str):
        self.status = status
        self.request_info = request_info
        self.message = message

    def __str__(self) -> str:
        return "{}, message={!r}, url={!r}".format(
            self.status,
            self.message,
            self.request_info.real_url,
        )

class EndpointDownError(Exception):
    """Endpoint is not alive (red flag)."""

    def __init__(self, endpoint_key: str) -> None:
        self.key = endpoint_key

    def __str__(self) -> str:
        return f"Endpoint {self.key} is not available."



import os
import sqlite3
import time
import yaml
from dataclasses import dataclass

from esi_requests.log import getLogger

logger = getLogger(__name__)


@dataclass(repr=False)
class CMDInfo:
    """Stores statistics for a database keyword, such as SELECT, DELETE, etc."""

    _cmd: str
    _cnt: int = 0
    _t: float = 0.0  # unit: nanosecond

    def __repr__(self) -> str:
        return f"(calls={self._cnt}, time={self._t / 1e9:.7f})"  # 7f because windows time has max 1e-07 resolution


@dataclass(init=False)
class _ESIDBStats:
    """Stores statistics for a database instance."""

    def __init__(self, db_name: str) -> None:
        self.db_name = db_name
        self.calls: int = 0

    def increment(self, cmd: str, _t: float):
        """Increments statistics of the database instance.

        Args:
            cmd: str
                A SQL keyword, such as SELECT or INSERT.
            _t: float
                Time spent for this db command. Unit in nanosecond.
        """
        self.calls += 1

        # If SQL query involves transaction, stored procedure, etc., they probably start with "BEGIN".
        # This is beyond the scope of this class design. You might see something like BEGIN=(calls=1, time=xxx).
        info: CMDInfo = getattr(
            self, cmd, CMDInfo(cmd)
        )  # cmd_info = self.SELECT || CMDInfo("SELECT")
        info._cnt += 1
        info._t += _t
        if not hasattr(self, cmd):
            setattr(self, cmd, info)

    def __repr__(self) -> str:
        nodef_f_vals = ((f, getattr(self, f)) for f in self.__dict__)

        nodef_f_repr = ", ".join(f"{name}={value}" for name, value in nodef_f_vals)
        return f"{self.__class__.__name__}({nodef_f_repr})"


class ESIDBManager:
    """Manages an SQLite database.

    Parameters:
        db_name (str): name of the database. The database file will be named as ``db_name.db``.
    """

    def __init__(self, db_name):
        self.db_name = db_name

        self.db_path = os.path.join(os.path.dirname(__file__), db_name + ".db")
        self.conn = sqlite3.connect(self.db_path)  # can't use isolation_level=None
        self.cursor = self.conn.cursor()

        self.__init_tables()
        self.__init_stats()  # self.tables
        logger.info("Database %s initiated at path %s", db_name, self.db_path)

    def __del__(self):
        self.close()

    @property
    def stats(self) -> _ESIDBStats:
        """Get statistics on database query performance."""
        return self._stats

    def execute(self, __sql: str, __parameters=...) -> sqlite3.Cursor:
        """Executes an SQL statement.
        
        Wraps cursor.execute with custom add-ons.
        Usage is the same (or should be the same) as cursor.execute() method of sqlite3.Cursor class.
        """
        cmd = __sql.split()[0]
        _s = time.perf_counter_ns()  # perf_counter has 1e-07 resolution in win32, lowest in time methods
        if __parameters is Ellipsis:
            cursor = self.cursor.execute(__sql)
        else:
            cursor = self.cursor.execute(__sql, __parameters)
        _t = time.perf_counter_ns() - _s
        self._stats.increment(cmd, _t)
        return cursor

    def commit(self) -> None:
        """Same as connection.commit() from sqlite3.Connection class."""
        self.conn.commit()

    def prepare_table(self, table_name: str, schema: str):
        """Prepares a table by creating it if not exists."""
        if table_name not in self.tables:
            # Read schema from schema.yml
            with open(os.path.join(os.path.dirname(__file__), "schema.yml")) as f:
                schema = yaml.full_load(f).get(schema)
                if schema is None:
                    raise ValueError(f"Schema {schema} not found in schema.yml")

            self.cursor.execute(f"CREATE TABLE IF NOT EXISTS {table_name} ({schema});")
            self.tables.append(table_name)
        else:
            logger.debug("Database %s already has table \"%s\"", self.db_name, table_name)

    def clear_table(self, table_name: str):
        """Clears a table using ``DELETE FROM table_name;``"""
        self.cursor.execute(f"DELETE FROM {table_name};")
        self.conn.commit()
        logger.debug("Clear table %s-%s successful", self.db_name, table_name)

    def drop_table(self, table_name: str):
        """Drops a table using ``DROP TABLE table_name;``"""
        self.cursor.execute(f"DROP TABLE IF EXISTS {table_name};")
        self.conn.commit()
        logger.debug("Drop table %s-%s successful", self.db_name, table_name)

    def clear_db(self):
        """Clears every table by calling ``clear_table()``."""
        for table in self.tables:
            self.clear_table(table)
        logger.debug("Clear DB %s successful", self.db_name)

    def close(self):
        """Close the database connection and cursor."""
        self.cursor.close()
        self.conn.close()
    
    def __init_tables(self):
        # with open(os.path.join(os.path.dirname(__file__), "schema.yml")) as f:
        #     dbconfig = yaml.full_load(f)
        # self._dbconfig = dbconfig.get(self.schema)
        # self.tables = self._dbconfig.get("tables")
        # for table in self.tables:
        #     table_config = self._dbconfig.get(table)
        #     schema = table_config.get("schema")
        #     self.cursor.execute(f"CREATE TABLE IF NOT EXISTS {table} ({schema});")
        
        # Get existing table names
        cursor = self.cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
        self.tables = [t[0] for t in cursor.fetchall()]

    def __init_stats(self):
        self._stats: _ESIDBStats = _ESIDBStats(self.db_name)



"""
esi_requests.models
~~~~~~~~~~~~~~~~~~~~

This module defines the data models used in esi_requests.

Models:
    - PreparedESIRequest: A prepared request object, used to send a request.
    - ESIResponse: The response object returned from an API call.

These models should not be created directly by users, but rather are returned by the `Session` object's request methods.

Some docstrings and code generated by ChatGPT, a large language model trained by OpenAI.

(c) 2023 by Hanbo Guo
"""

import json
from dataclasses import dataclass
from typing import Dict, Optional

from .exceptions import ESIResponseError


class ESIRequest:
    """Holds user-provided parameters for a request to ESI.

    An ESIRequest instance is created with parameters like the request endpoint,
    HTTP method, query parameters, and request headers. This instance is then passed to
    the `prepare_request()` method in the `Session` class to create a `PreparedESIRequest`
    instance which can be sent as an HTTP request to the ESI server via `send()` method.

    Attributes:
        endpoint (str): An endpoint of EVE ESI.
        method (str): The HTTP method for the request.
        params (dict): The query parameters to include in the request.
        headers (dict): The headers for the request.
        prepared (bool): Whether the request is prepared or not.
    """

    def __init__(self, endpoint: str, method: str, params: Dict, headers: Dict, **kwargs) -> None:
        self.endpoint = endpoint
        self.method = method
        self.params = params
        self.headers = headers
        self.kwargs = kwargs

        self.prepared = False


class PreparedESIRequest:
    """
    The prepared request object with necessary details to be sent to ESI server. 

    Note:
        User should not create ``PreparedESIRequest`` but gets it from API calls via ``ESISession.prepare_request()`` method.
    """
    def __init__(self, endpoint: str = None, method: str = None, url: str = None, params: Dict = None,
                 headers: Dict = None, kwargs: Dict = None):
        """
        Args:
            endpoint (str): The endpoint for the request.
            method (str): The HTTP method for the request (e.g. GET, POST).
            url (str): The full URL for the request, including the endpoint and any query parameters.
            params (dict): A dictionary containing any query parameters to include in the request.
            headers (dict): A dictionary containing any headers to include in the request.
            kwargs (dict): Any additional arguments to be passed to the request function.
        """
        self.endpoint: str = endpoint
        self.method: str = method
        self.url: str = url
        self.params: Dict = params or {}
        self.headers: Dict = headers or {}
        self.kwargs: Dict = kwargs or {}

    def __repr__(self) -> str:
        return f"<PreparedESIRequest [{self.method}]>"


@dataclass
class ESIResponse:
    """The response object returned by ESI API calls.

    Attributes:
        status (int): The HTTP status code of the response.
        reason (Optional[str]): The reason-phrase from aiohttp.ClientResponse.
        headers (dict): A case-insensitive dictionary of HTTP response headers.
        url (str): The URL of the request.
        request_info (ESIRequest): A reference to the `PreparedESIRequest` object.
        text (Optional[str]): The response body, decoded using utf-8.
        expires (Optional[str]): A RFC7231 formatted datetime string, if any.

    Note:
        Do not create an `ESIResponse` object directly, but retrieve it from ESI API calls.

        ESI API uses `utf-8` encoding for response bodies.

    Methods:
        __bool__(): Returns `True` if `status` is `200` or `304`, `False` otherwise.
        __repr__(): Returns a string representation of the `ESIResponse` object.
        ok: Returns `True` if `status` is `200` or `304`, `False` otherwise.
        raise_for_status(): Raises an `ESIResponseError` if `status` is greater than or equal to 400.
        json(): Returns the JSON-deserialized response body.
    """


    status: int = None  # e.g. 404, 200
    reason: Optional[str] = None
    headers: Dict = None
    url: str = None
    request_info: "ESIRequest" = None
    text: Optional[str] = None
    expires: Optional[str] = None

    def __bool__(self) -> bool:
        return self.ok

    def __repr__(self) -> str:
        return f"<Response [{self.status}]>"

    @property
    def ok(self) -> bool:
        """Returns ``True`` if ``status`` is ``200`` or ``304``, ``False`` if not.

        This is **not** a check for ``200 OK``.
        """
        return self.status in (200, 304)

    def raise_for_status(self):
        if self.status >= 400:
            raise ESIResponseError(self.status, self.request_info, self.reason)

    def json(self):
        """Returns json deserialized response body.

        Note that ESI all uses ``utf-8`` for encoding.
        """
        return json.loads(self.text)



from dataclasses import dataclass
from itertools import product
from typing import Any, List, Optional

from .__version__ import __version__


def default_headers():
    return {
        "User-Agent": f"python-esi_requests/{__version__}",
    }


def cross_product(**kwargs):
    keys = kwargs.keys()
    vals = kwargs.values()
    for instance in product(*vals):
        yield dict(zip(keys, instance))


@dataclass
class Param:
    """Hold a Param for an API

    Attributes:
        name: A str for the name of the param. E.g. "character_id"
        _in: A str for the occurance of the param. One of ["path", "query", "header"]
        required: A bool for requirement.
        dtype: A str for data type of the param.
        default: Default value for the Param.
    """

    name: str  # characther_id
    _in: str  # "path" / "query" / "header" / "body"
    required: bool
    dtype: type  # "string" / "integer" / "array" / "boolean" / "" (schema)
    default: Optional[Any] = None


@dataclass
class ESIParams:
    """A list like datastructure of Param(s).

    Attributes:
        params: A list of Param instance.
    """

    # list of ESI pre-defined meta parameters, initialized at the start
    # Param.name: #/parameters/{actual_name}
    params: List[Param]

    def append(self, param: Param) -> None:
        """Append a Param similar to list append.
        No value checking. Used for testing.
        """
        self.params.append(param)

    def __getitem__(self, name: str) -> Param:
        """Returns a reference of Param with name.
        If no Param with name found, return None
        """
        for p in self.params:
            if p.name == name:
                return p

        return None

    def __contains__(self, name: str) -> bool:
        """Returns True if Param with ``name`` exists."""
        for p in self.params:
            if p.name == name:
                return True

        return False

    def __iter__(self):
        yield from self.params



"""Validates a given JWT token originating from the EVE SSO.

Source: https://github.com/esi/esi-docs/blob/master/examples/python/sso/validate_jwt.py
"""
import requests
import sys
from jose import jwt
from jose.exceptions import ExpiredSignatureError, JWTError

from esi_requests.log import getLogger

logger = getLogger(__name__)


def validate_eve_jwt(jwt_token) -> dict:
    """Validate a JWT token retrieved from the EVE SSO.

    Args:
        jwt_token: A JWT token originating from the EVE SSO
    Returns
        dict: The contents of the validated JWT token if there are no
              validation errors
    """

    jwk_set_url = "https://login.eveonline.com/oauth/jwks"

    res = requests.get(jwk_set_url)
    res.raise_for_status()

    data = res.json()

    try:
        jwk_sets = data["keys"]
    except KeyError as e:
        logger.error(
            "The returned JTW payload did not have the expected key {}. "
            "Payload returned from the SSO looks like: {}".format(e, data)
        )
        sys.exit(1)

    jwk_set = next((item for item in jwk_sets if item["alg"] == "RS256"))

    try:
        return jwt.decode(
            jwt_token,
            jwk_set,
            algorithms=jwk_set["alg"],
            issuer="login.eveonline.com",
            audience="EVE Online",  # newly added required field
        )
    except ExpiredSignatureError:
        logger.error("The JWT token has expired")
        sys.exit(1)
    except JWTError as e:
        logger.error(f"The JWT signature was invalid: {e}")
        sys.exit(1)



import json
import os
from dataclasses import dataclass
from typing import List

import requests

from esi_requests.log import getLogger

from .utils import ESIParams, Param

logger = getLogger(__name__)


@dataclass
class EndpointMetadata:
    endpoint: str = None

    # List of methods that this endpoint supports: ["get", "head", ...]
    supported_methods: List[str] = None

    # List of parameters accepted by the endpoint
    parameters: ESIParams = None

    # Required OAuth scope for this endpoint
    sso_scope: List[str] = None


class ESIMetadata:
    """Parses Swagger matadata of ESI."""

    PARAM_DEFAULT = {"order_type": "all"}
    METADATA_PATH = os.path.join(os.path.dirname(__file__), "metadata.json")

    def __init__(self):
        # Swagger field: defines endpoints
        self.paths = None

        # Swagger field: defines OAuth scopes
        self.securityDefinitions = None

        # Swagger field: defines accepted parameters
        self.parameters = None

        self.__load_metadata()

    def __call__(self, endpoint: str) -> EndpointMetadata:
        """Extract information of an endpoint from Swagger metadata.

        Parse and find metadata entry with the given key.
        Assume each metadata entry is ONE of GET, POST.
        Some ESI API has DELETE or PUT, such as /characters/{character_id}/contacts/,
        but they tend to be trivial in market data analysis, so they are not supported.

        Returns:
            An ESIRequest instance with request_key = key.

        Raises:
            KeyError: key is not a valid request type.
        """
        if not endpoint in self.paths.keys():
            raise KeyError(f"{endpoint} is not a valid request type.")

        accepted_methods = list(self.paths[endpoint].keys())
        
        # Some endpoints have more than one accepted request methods,
        # and each request methods have its own parameters and sso scopes requirement.
        # But most useful endpoints have only one request method.
        # For simplicity, only take the information of the first request method.

        parameters = self.parse_parameters(endpoint)
        security = self.parse_security(endpoint)

        return EndpointMetadata(endpoint, accepted_methods, parameters, security)

    def parse_parameters(self, endpoint: str) -> ESIParams:
        """Extract what parameters are accepted for an endpoint.

        Returns:
            An ESIParams instance containing all accepted parameters for the endpoint.
        """
        accepted_methods = list(self.paths[endpoint].keys())
        body = self.paths[endpoint][accepted_methods[0]]

        parameters = body["parameters"]
        params = []
        for param in parameters:
            # check for {$ref : #/parameters/xxx} type param
            # Parameters defined in metadata["parameters"] has key pattern: "$ref/parameters/xxx".
            # Ignore parameters with $ref/parameters signature but not in metadata["parameters"] field.
            metaparam = param.get("$ref", "")  # $ref for meta parameters
            if metaparam:
                param_ = self.parameters[metaparam.split("/")[-1]]
                if param_:
                    params.append(param_)
                continue

            # Construct Param class
            # Param.default is only present in meta parameters.
            # Some params should have default value but absent in meta parameters,
            # which are further defined in PARAM_DEFAULT dictionary.
            params.append(
                Param(
                    name=param["name"],
                    _in=param["in"],
                    required=param.get("required", False),
                    dtype=param.get("type", ""),
                    default=self.PARAM_DEFAULT.get(param["name"]),
                )
            )

        return ESIParams(params)

    def parse_security(self, endpoint: str) -> List[str]:
        """Extra what OAuth scopes are needed for an endpoint.

        Each API request have either 0 or 1 scope required.

        Note:
            API request with multiple scopes will yield error in finding token in ``ESITokens``.

        Raises:
            ValueError: API request with multiple scopes is not supported.
        """
        accepted_methods = list(self.paths[endpoint].keys())
        body = self.paths[endpoint][accepted_methods[0]]

        security = body.get("security", None)

        if not security:
            # this endpoint does not require any scopes
            return []

        if len(security) > 1:
            # Only one OAuth method accepted: "evesso"
            # Maybe in the future this block will be reached.
            pass

        scope_ = security[0]["evesso"]
        if len(scope_) > 1:
            raise ValueError(f"endpoints with > 1 scopes are not supported.")
            
        return scope_

    def __load_metadata(self) -> None:
        """Loads Swagger metadata into this object.

        Raises:
            ValueError: Metadata is empty when loading from local file or EVE website.
        """
        metadata = None

        # TODO: retrieve from ESI every three(?) months
        if (
            not os.path.exists(self.METADATA_PATH)
            or os.stat(self.METADATA_PATH).st_size == 0
        ):
            # If no local copy, retrieve from ESI
            r = requests.get(
                "https://esi.evetech.net/latest/swagger.json?datasource=tranquility"
            )
            r.raise_for_status()
            metadata = r.json()

            with open(self.METADATA_PATH, "w") as metadata_file:
                json.dump(metadata, metadata_file)
                logger.debug("metadata retrieved from ESI")
        else:
            # If local copy exists, just read
            with open(self.METADATA_PATH, "r") as metadata_file:
                metadata = json.load(metadata_file)

        if not metadata or not metadata.keys():
            raise ValueError("Metadata is empty.")
        
        self.securityDefinitions = metadata["securityDefinitions"]
        self.paths = metadata["paths"]

        # Store metadata parameters in ESIParams
        params = [
            Param(
                name=v['name'],                         # param name
                _in=v["in"],                            # in headers, query, or path
                required=v.get("required", False),
                dtype=v["type"],                        # accepted data type
                default=v.get("default", None),         # default value if not given by user
            )
            for v in metadata["parameters"].values()
        ]
        self.parameters = ESIParams(params)



import hashlib
import inspect
import os
import pickle
import re
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Callable, Coroutine, Dict, List, Tuple, Union

from esi_requests.log import getLogger

logger = getLogger(__name__)

if TYPE_CHECKING:
    from .db import ESIDBManager


# ---- Cache Utility classes ---- #


class _CacheRecordBaseClass:
    """Base class holding all cache instances.
    Provides stats over all instances."""

    instances = set()

    @property
    def record(self):
        """Returns a list of record, one for each cache instance."""
        ret = [_cache.record for _cache in self.instances]
        return ret


CacheStats = _CacheRecordBaseClass()


@dataclass
class _CacheRecord:
    """Records hits and misses of a cache instance."""

    db_name: str
    table: str
    caller: str = None
    hits: int = 0
    miss: int = 0


class InsertBuffer:
    """Buffers cache.set to avoid repetitive insert transaction.
    
    Note:
        This buffer might contain expired database entries, 
        but it gurantees to return the **newest** entry of a given key.
        Also, upon flushing to database, 
        this buffer gurantees only the **newest** entry exists in the database after a flush.
    """

    def __init__(self, db: "ESIDBManager", cap: int = 50) -> None:
        self.db = db
    
        self.buffer: List[Tuple] = []  
        # [(
        #   (key_hash, value, expires), table
        # ), (
        #   (key_hash, value, expires), table
        # )...]
        self.cap = cap

    def flush(self) -> None:
        """Flushes buffer payload to database file. Buffer payload is cleared after flushing."""
        self.db.execute("BEGIN")
        for entry in self.buffer:
            self.db.execute(f"INSERT OR REPLACE INTO {entry[1]} VALUES({','.join('?'*len(entry[0]))})", entry[0])
        self.db.commit()
        self.clear()
        logger.debug("Insertbuffer entries flushed")

    def insert(self, entry: Tuple, table: str) -> None:
        """Inserts db entry to buffer. Flushes if ``cap`` is reached.

        Args:
            entry: (key, value, expires)
                A database entry.
            table: str
                Insert entry to this table.
        """
        if len(self.buffer) >= self.cap:
            self.flush()

        self.buffer.append((entry, table))

    def select(self, key_hash) -> Tuple:
        """Selects value from buffer. Similar to cache.get.

        Args:
            key_hash: hash(key)
                A hashed key, which is retrieved from hash_key() function.
        """
        # Reverse order ensures the newest entry of the key_hash is returned
        for i in reversed(range(len(self.buffer))):
            if self.buffer[i][0][0] == key_hash:
                return self.buffer[i][0]
        return None

    def clear(self) -> None:
        """Clears buffer paylaod. Resets to empty list."""
        del self.buffer
        self.buffer = []

    def __contains__(self, key):
        _h = hash_key(key)
        for b in self.buffer:
            if b[0][0] == _h:
                return True
        return False

    def __len__(self):
        return len(self.buffer)


class srcodeBuffer:
    """Buffers inspect.getsource, which takes ~10s out of ~100s in a market history request with 10k type_ids."""

    payload: Dict = {}  # not a dataclass

    @classmethod
    def getsource(cls, func: Union[Callable, Coroutine]):
        key = func.__qualname__
        srcode = cls.payload.get(key)
        if srcode is None:
            srcode = inspect.getsource(func)
            cls.payload[key] = srcode
        return srcode


class _DeleteHandler:
    """Tracks ``expires`` time and deletes when the time has come.
    
    This handler holds a list of datetimes,
    at which this handler deletes **ALL** cache lines that has expired.

    This handler only gurantees to delete expired cache lines no later than 5 minutes passed the expired time.
    It does not delete expired cache lines immediately at the ``expires`` timestamp.
    """

    def __init__(self, db: "ESIDBManager", table: str):
        self.db = db
        self.table = table

        self.schedule: List = []  # priority queue

        # Read from local cache
        db_dir = os.path.join(os.path.dirname(__file__), ".deleter")
        if not os.path.exists(db_dir):
            # Race condition
            try:
                os.makedirs(db_dir)
            except FileExistsError:
                pass
        self.path = os.path.realpath(os.path.join(db_dir, f"{db.db_name}-{table}-delete_schedule.tmp"))
        if os.path.exists(self.path) and os.stat(self.path).st_size > 0:
            with open(self.path, "rb") as f:
                self.schedule = sorted(pickle.load(f))

        self.last_delete: datetime = None

    def update(self, expire: datetime) -> None:
        """Adds a new expire time to keep track on, and deletes if any delete time has passed.

        Args:
            expire: datetime
                A new expire time.
        """
        # Seperate two deletes by 5 minutes
        minute = expire.minute
        expire = expire.replace(minute=int(minute / 5) * 5, second=0, microsecond=0)
        expire += timedelta(minutes=5)  # round up 5 minutes

        if expire not in self.schedule:
            self.schedule.append(expire)
            self.schedule.sort()  # good enough for small list

        # Find the latest time that triggers a delete
        now = datetime.utcnow()
        latest_expire = None
        for i in range(len(self.schedule)):
            t = self.schedule[i]
            if t < now:
                latest_expire = t
            else:
                break

        if latest_expire is not None:
            # Only deletes from db.
            # If InsertBuffer entries expired, they will be dealt in later runs.
            self.db.execute(f"DELETE FROM {self.table} WHERE expires < ?", (latest_expire,))
            self.db.commit()
            logger.debug("Cache DELETE attempted")
            self.last_delete = datetime.utcnow()
            for _ in range(i):  # remove all time that's passed
                self.schedule.pop(0)

    def save(self) -> None:
        """Saves payload to local tmp file, serialize using pickle."""
        with open(self.path, "wb") as f:
            pickle.dump(self.schedule, f)


# ---- Utility functions ---- #


def make_cache_key(func: Union[Callable, Coroutine], *args, **kwd):
    """Hashes a function and its arguments using sha256.

    The function is hashed using function_hash() to a string.
    The args and kwds are pickled to bytes.
    If any of args or kwd is a function, this function encodes them by inspecting source code.
    If the source code is changed, this function encodes a different value.

    Args:
        func: A function or a coroutine.
        args: Argument passed to the api function in argument form.
        kwd: Keyword arguments passed to the api function.

    Returns:
        A set containing (function_hash, str(args), str(kwd))

    Examples:
    >>> # Encode: get_market_history("The Forge", reduces=reduce_volume)
    >>> key_before = make_cache_key(get_market_history, "The Forge", reduce_volume)
    >>> # If reduce_volume is changed, even if a space or an extra line is added
    >>> key_after = make_cache_key(get_market_history, "The Forge", reduce_volume)
    >>> assert key_before != key_after
    """
    func_args = list(args)
    func_kwd = kwd.copy()
    for i in range(len(func_args)):
        if isinstance(func_args[i], Callable):
            func_args[i] = function_hash(func_args[i])
        if isinstance(func_args[i], list):
            func_args[i] = list(set(func_args[i]))
    for k in func_kwd:
        if isinstance(func_kwd[k], Callable):
            func_kwd[k] = function_hash(func_kwd[k])
        if isinstance(func_kwd[k], list):
            func_kwd[k] = list(set(func_kwd[k]))
    _h = function_hash(func)

    ret = (_h, pickle.dumps(func_args), pickle.dumps(func_kwd), func.__qualname__)
    return ret


def function_hash(func: Union[Callable, Coroutine]):
    """Hashes a function.

    Hashes a function based on its source code. If the source code is modified in any way,
    this function hashes to a different value. If docstring is changed, hash should remain intact.
    """
    source_code = srcodeBuffer.getsource(func)
    source_code = re.sub(r'"""[\w\W]+?"""\n', "", source_code)  # remove docstring
    return "esi_function-{}-{}".format(
        func.__qualname__,
        hashlib.sha256(source_code.encode("utf-8")).hexdigest(),
    )


def hash_key(key) -> str:
    """Default hashing function for a key. Using sha256 as hash function."""
    return hashlib.sha256(pickle.dumps(key)).hexdigest()



import json
import os
from functools import wraps
from time import time
from typing import TYPE_CHECKING, Coroutine, Dict

import aiohttp
import pandas as pd
import requests

from .data import SqliteCache
from .data.utils import make_cache_key
from .exceptions import EndpointDownError, InvalidRequestError
from .log import getLogger
from .metadata import ESIMetadata
from .models import ESIRequest

if TYPE_CHECKING:
    from .models import PreparedESIRequest, ESIResponse
    

logger = getLogger(__name__)


def cache_check_request(func: Coroutine):
    """Caches various parameter checkings, such as check_type_id."""
    # func has signature: async def check_{xxx_id}(self, request) -> bool
    @wraps(func)
    async def cache_check_request_wrapped(_self: "ESIRequestChecker", *args, **kwd):
        # Caches _RequestChecker methods
        key = make_cache_key(func, *args, **kwd)
        value = _self.cache.get(key)
        if value is not None:  # cache hit
            return value

        ret = await func(_self, *args, **kwd)  # exec

        expires = 24 * 3600 * 30  # one month
        _self.cache.set(key, ret, expires)
        return ret

    return cache_check_request_wrapped


class _NonOverridable(type):
    """Prevents subclass overriding some methods."""

    __final__ = ["__call__", "__check_request"]  # methods not overridable

    def __new__(cls, __name: str, __bases, __namespace):
        if __bases:
            for finals in cls.__final__:
                if finals in __namespace:
                    raise SyntaxError(f"Overriding {finals} is not allowed")
        return type.__new__(cls, __name, __bases, __namespace)


class ESIRequestChecker(metaclass=_NonOverridable):
    """Checks if a request would cause an error on ESI side.

    Rule-based parameter checking to avoid errors from ESI.
    The goal is to completely eliminate 400 and 404 errors in stable state.

    Note:
        Individual check methods should be async functions, and could be decorated by ``cache_check_request``.
        User could override individual check methods to customize checking rules.
        ``__call__`` and ``__check_request`` methods are not allowed to override.
    """
    # It seems like if the parameter to be checked is in path (appears as {xxx_id} in endpoint name),
    # incorrect value would cause a 404 error.
    # Instead, if the parameter is in query, an empty response body would be given (probably database select returns nothing).

    # TODO: give priorities to different checks
    ## e.g. esi_requests.get("/markets/{region_id}/order/", region_id=123, type_id=[-1])
    ## should check region_id first, then type_id.
    # TODO: different response generation rules
    ## e.g. esi_requests.get("/universe/types/{type_id}/", type_id=-1) -> 404
    ## e.g. esi_requests.get("/markets/{region_id}/order/", region_id=10000002, type_id=[-1]) -> 200, empty response body
    def __init__(self) -> None:
        self.enabled = True
        self.raise_flag = False

        self.endpoints_checker = ESIEndpointChecker()
        self.metadata_parser = ESIMetadata()

        self.cache = SqliteCache("request_cache", "request_checker")
        self.invTypes = self.__load_sde()

        self._requests_made = 0  # just for fun
        self.__async_session = None

    async def __call__(self, request: "PreparedESIRequest", raise_flag: bool = False) -> bool:
        if not self.enabled:
            return True
        
        if self.__async_session is None:
            self.__async_session = aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=False))

        self.raise_flag = raise_flag
        return await self.__check_request(request)

    async def __check_request(self, request: "PreparedESIRequest") -> bool:
        """Checks if an ESIRequest is valid.

        Checks parameters of an ESIRequest, and predicts if the request is valid.
        Currently, the ESI._check_* family only checks parameters following some rules.
        This means there is no feedback loop from responses.

        Raises:
            InvalidRequestError: raised when request is blocked and ESI.request family sets keyword ``raises = True``.
            EndpointDownError: raised when requested endpoint is down.

        Note:
            This method is not cached, but individual checks might be cached for one month.
        """
        valid = True
        error = None

        metadata = self.metadata_parser(request.endpoint)

        # Check type_id in query
        if valid:
            if "type_id" in request.kwargs:
                type_id = request.kwargs.get("type_id")
            elif "type_id" in request.params:
                type_id = request.params.get("type_id")
            else:
                type_id = None

            # If "type_id" not required by this endpoint, just ignore checking
            if "type_id" in metadata.parameters:
                if type_id is None and not metadata.parameters["type_id"].required:
                    # sometimes type_id = None is valid, so no check
                    valid = True
                else:  # check
                    valid = await self.check_type_id(type_id)
            if not valid:
                error = InvalidRequestError("type_id", type_id)

        # Raise error if necessary
        if not valid:
            self.__log(request)
            # request.blocked = True
            if self.raise_flag is True and error is not None:
                raise error from None
            else:
                return self.raise_flag

        return valid

    @cache_check_request
    async def check_type_id(self, type_id: int) -> bool:
        """Checks if a type_id is valid.

        First checks using SDE, then checks using ESI endpoint.

        Note:
            This method is cached for one month.
        """
        valid = type_id in self.invTypes["typeID"].values

        if valid is True:
            invType = self.invTypes.loc[self.invTypes["typeID"] == type_id]
            valid = bool(int(invType["published"]))

        if valid is True:
            success = False
            attempts = 3
            while not success and attempts > 0:
                async with self.__async_session.get(
                    f"https://esi.evetech.net/latest/universe/types/{type_id}/?datasource=tranquility&language=en",
                ) as resp:
                    if resp.status == 502:
                        attempts -= 1
                        continue
                    if resp.status == 200:
                        success = True
                    data: dict = await resp.json()
                    self._requests_made += 1
                    valid = data.get("published")

        return valid
    
    async def close(self):
        if self.__async_session is not None:
            await self.__async_session.close()
    
    def __load_sde(self) -> pd.DataFrame:
        # Reading a .csv.bz2 is costly. Takes 15MB memory and a long time (~0.x second)
        # Retrieve from Fuzzwork if local copy not exists
        static_path = os.path.join(os.path.dirname(__file__), "data", "static")
        if not os.path.isdir(static_path):
            try:
                os.mkdir(static_path)
            except FileExistsError:
                pass
        invTypes_path = os.path.join(static_path, "invTypes.csv.bz2")
        if not os.path.exists(invTypes_path):
            resp = requests.get("https://www.fuzzwork.co.uk/dump/latest/invTypes.csv.bz2")
            with open(invTypes_path, "wb") as f:
                f.write(resp.content)
            
        return pd.read_csv(invTypes_path)

    def __log(self, api_request: ESIRequest):
        logger.warning(
            'BLOCKED - endpoint_"%s": %s',
            api_request.endpoint,
            api_request.kwargs,
        )


class ESIEndpointChecker:
    """Checks status of an ESI endpoint.

    Note:
        This method does not follow the ``expires`` field in response header.
        This method retrieves ``status.json`` from ESI every 60 seconds (as ``expires`` headers specified).
    """

    def __init__(self) -> None:
        self.enabled = True
        self.target_url = "https://esi.evetech.net/status.json?version=latest"
        self.fd_path = os.path.join(os.path.dirname(__file__), "data", "status.json")

        if not os.path.exists(self.fd_path) or os.stat(self.fd_path).st_size == 0:
            self.fd = open(self.fd_path, "w")
            self.status_parsed = None
        else:
            self.fd = open(self.fd_path, "r")
            self.status_parsed = json.load(self.fd)

    def __del__(self):
        self.fd.close()

    @property
    def fd_expired(self) -> bool:
        return (self.status_parsed is None or len(self.status_parsed) == 0) or (
            os.path.exists(self.fd_path) and os.path.getmtime(self.fd_path) - time() > 60
        )  # ESI server's status.json expires every 60 seconds

    def check(self, endpoint: str) -> bool:
        """Checks if an endpoint is available."""
        # If no local status.json, or local version expired, retrieve from ESI
        if self.fd_expired:
            resp = requests.get(self.target_url)
            status = resp.json()
            self.status_parsed = self._parse_status_json(status)
            json.dump(self.status_parsed, self.fd)
            self.fd.flush()

        # Now, self.status_parsed has a fresh **parsed** copy of ``status.json``
        return self.status_parsed.get(endpoint, False)

    @staticmethod
    def _parse_status_json(status) -> Dict:
        return {entry["route"]: True if entry["status"] == "green" else False for entry in status}


class FakeResponseGenerator:

    def __init__(self):
        self.cache = SqliteCache("request_cache", "resp_generator")

        self.__async_session = None

    def __call__(self, *args, **kwargs):
        return self.generate()

    def generate(self, request: "PreparedESIRequest") -> "ESIResponse":
        """Generates a fake response for a given PreparedESIRequest.

        Returns:
            ESIResponse: a fake response for a given request.
        """
        self.cache.get(request.endpoint)
        return ESIResponse()



"""
esi_requests.parser
~~~~~~~~~~~~~~~~~~~

This module converts user-given parameters to an ESIRequest.
"""

from copy import deepcopy
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional

from .log import getLogger
from .metadata import ESIMetadata
from .models import PreparedESIRequest
from .sso.application import ESIApplications
from .sso.token import ESITokens
from .utils import cross_product, default_headers

if TYPE_CHECKING:
    from .metadata import EndpointMetadata
    from .models import ESIRequest
    from .utils import Param

logger = getLogger(__name__)


@dataclass
class ETagEntry:
    """Struct for storing etag and etag's payload."""

    etag: str
    payload: Optional[Any]


class ESIRequestParser:
    """Process an ``ESIRequest`` object to a ``PreparedESIRequest`` object."""

    SUPPORTED_METHODS = ["get", "head"]
    HOST = "https://esi.evetech.net/latest"  # also available on metadata

    def __init__(self) -> None:
        self.metadata_parser = ESIMetadata()
        self.apps = ESIApplications()

    def __call__(self, request: "ESIRequest") -> List["PreparedESIRequest"]:
        """Parses user input ``key``, ``method``, and keywords to a ``ESIRequest``.

        This method first consults with ``ESIMetadata`` to retrieve metadata information of a request endpoint.
        Then fills users input parameters to ``headers`` and ``params``.

        Args:
            key: str
                ESI request key, in "/.../.../" format, retrieved from ESI website.
            method: str
                One of ``get``, ``head``. ``post`` and other methods are not supported.

        Returns:
            An ``ESIRequest`` instance, with all relevent fields filled in.
        """
        metadata = self.metadata_parser(request.endpoint)  # metadata of endpoint

        # Make sure the user is using a correct request method
        self.__check_method(metadata, request.method)

        requests = self.parse(request, metadata)

        return requests
    
    def parse(self, request: "ESIRequest", metadata: "EndpointMetadata") -> List["PreparedESIRequest"]:
        if request.params is None:
            request.params = {}
        if request.headers is None:
            request.headers = {}

        # Unpack parameters, such as if given a request with (type_id=[1,2,3], region_id=[10, 11]),
        # we need to treat it as six seperate request: (type_id=1, region_id=10), (type_id=1, region_id=11), ...
        loop_through = self.__find_what_to_loop(request)
        loop_cross_product = cross_product(**loop_through)

        # Remove parameters in ``loop_through`` from request
        for param in loop_through:
            if param in request.headers:
                del request.headers[param]
            elif param in request.params:
                del request.params[param]
            elif param in request.kwargs:
                del request.kwargs[param]

        # Prepare shared headers, kwargs, and params
        headers = default_headers()
        headers.update(request.headers)
        # Add oauth to headers
        if metadata.sso_scope and "Authorization" not in headers:  # some method does not need oauth
            app = self.apps.search_scope(" ".join(metadata.sso_scope))  # find matching application
            with ESITokens(app) as tokens:  # find token for this scope
                cname = request.kwargs.pop("cname", "any")  # "any" for matching any token
                token = tokens.generate() if not tokens.exist(cname) else tokens[cname]
                headers["Authorization"] = "Bearer {}".format(token.access_token)
                if "character_id" not in request.kwargs:
                    # Tackle ESI's authenticated search endpoint, see https://github.com/esi/esi-issues/issues/1323.
                    # Use character_id of one of the authenticated user if character_id is not given by user.
                    request.kwargs["character_id"] = token.character_id
        request.headers = headers
  
        # Construct the list of PreparedESIRequest
        requests = []
        for loop_body in loop_cross_product:
            prep = PreparedESIRequest(endpoint=request.endpoint, method=request.method, params=deepcopy(request.params), headers=deepcopy(request.headers), kwargs=deepcopy(request.kwargs))
            
            # Since loop_body has same keys for all requests, this ``update`` could work.
            prep.kwargs.update(loop_body)

            self.parse_request_parameters(prep, metadata)

            self.build_req_headers(prep)
            self.build_req_url(prep)

            requests.append(prep)

        return requests

    def build_req_url(self, request: "PreparedESIRequest") -> None:
        """Builds url for a request from ``endpoint`` and ``params`` values.
        
        ``params`` should already be parsed and checked against ESI metadata.
        Use ``parse_request_keywords()`` before calling this function.
        """        
        params = [f"{k}={request.params[k]}" for k in request.params]
        path_param = request.kwargs.pop("path_params", None)

        url = self.HOST + request.endpoint.format(**path_param)
        url = f"{url}?{'&'.join(params)}"

        request.url = url

    def build_req_headers(self, request: "PreparedESIRequest") -> None:
        """Builds ``headers`` for a ``PreparedESIRequest``.

        Most ESI request needs no user-input headers, but some request headers are useful,
        such as ``If-None-Match`` and ``User-Agent``.
        This method defines default value of these useful HTTP request headers field.
        """

        # If-None-Match is updated in Session
        # User-Agent and other static headers are defined in ``default_headers()`` in utils

        pass

    def parse_request_parameters(self, request: "PreparedESIRequest", metadata: "EndpointMetadata") -> None:
        """Parses user input parameters according to endpoints metadata.

        Checks fields in ``keywords`` if necessary parameters are given.
        If given, fills in ``api_request`` with fields from keywords.
        If not, either raises errors if field is ``required``, or fill in default value if defined by ESI metadata.

        Args:
            api_request: ESIRequest
                Request info for a request. Retrieved from ``ESIMetadata.__getitem__``.
                Necessary info (url, params, headers) should be filled in by ``ESIMetadata``.
            keywords: dict
                Keywords provided by user, containing ``headers``, ``params``, and other necessary fields defined by ESI,
                such as ``character_id``, ``type_id``, etc.
                If ESI metadata marks fields as ``required``, missing these fields in ``keywords`` raises errors.

        Facts:
            _in: path
                1. Always marked as ``required``
                2. Appears as ``Param.name`` in url: https://.../characters/{character_id}/orders/
                3. Pass in with kwd argument, not params, headers, or data
                4. No default value defined by ESI
            _in: query
                1. Pass in with either ``kwd`` or ``params``, not ``headers``
                2. Appears as ``?query=value`` in url: https://.../?datasource=tranquility
            _in: header
                1. Request token has been updated to headers before calling this function
                2. ESI marks ``token`` field as optional

        Note:
            ``dtype`` field in ESI metadata is not checked nor enforced.
        """

        path_params = {}  # parameter for endpoint, such as region_id for /markets/{region_id}/orders/
        query_params = {}  # paramters for url/?{key1}={value1}?{key2}={value2}...
        kwargs = deepcopy(request.kwargs)

        for api_param_ in metadata.parameters:
            # Each param has a "in" field defined by metadata.
            if api_param_._in == "path":
                value = self.__parse_request_parameter_in_path(kwargs, api_param_)
                path_params.update({api_param_.name: value})

            elif api_param_._in == "query":
                default = api_param_.default
                value = self.__parse_request_parameter_in_query(kwargs, request.params, api_param_)
                if value is not None:
                    query_params.update({api_param_.name: value})  # update if value is given
                elif default is not None:
                    query_params.update({api_param_.name: default})  # else update if default is set

            elif api_param_._in == "header":  # not "headers"
                value = self.__parse_request_parameter_in_header(request.headers, api_param_)
                if value is not None:
                    request.headers.update({api_param_.name: value})

        request.kwargs.update({"path_params": path_params})
        request.params.update(query_params)

    @staticmethod
    def __parse_request_parameter_in_path(where: dict, param: "Param") -> str:
        # dtype is not checked yet. Checking it needs to parse "schema" field and integerate into "dtype",
        # and needs to find a way to fit user input to the dtype field.
        # No need to check Param.required because Param._in == "path" => Param.required == True
        value = where.pop(param.name, None)
        if value is None:
            raise ValueError(f'Required parameter "{param.name}" not given.')
        return value

    @staticmethod
    def __parse_request_parameter_in_query(kwargs: dict, params: dict, param: "Param") -> str:
        value = kwargs.pop(param.name, None)
        value2 = params.get(param.name)

        if value and value2:
            raise ValueError(f'Duplicated parameter "{param.name}" give in both params and kwargs.')
    
        if value is None and value2 is None and param.required and param.default is None:
            # If parameter not given, and no default value exists, but is required, raise error
            raise ValueError(f'Requred parameter "{param.name}" not given.')

        if value is None and value2 is None:
            return param.default
        if value is not None:
            return value
        if value2 is not None:
            return value2

    @staticmethod
    def __parse_request_parameter_in_header(headers: dict, param: "Param") -> str:
        if param.name not in headers:
            if param.required:
                raise KeyError(f'Required parameter "{param.name}" not given in keywords.')
            else:
                return param.default
    
    @staticmethod
    def __find_what_to_loop(request: "ESIRequest") -> Dict[str, List]:
        """Given a ``esi_requests.get("endpoint", type_id=[1,2,3,4,5], region_id=[2,3,4])``,
           this method captures ``"type_id"`` and ``"region_id"``."""
        kwargs = deepcopy(request.kwargs)
        if request.params:
            kwargs.update(request.params)  # combine kwargs and params
        if request.headers:
            kwargs.update(request.headers)  # combine with headers
        should_loop_or_not = lambda value: isinstance(value, Iterable) and not isinstance(value, str)

        loop_through = {param: kwargs[param] for param in kwargs if should_loop_or_not(kwargs[param])}

        return loop_through

    def __check_method(self, metadata: "EndpointMetadata", method: str) -> None:
        """Checks if request method is supported by ESI.

        Args:
            api_request: ESIRequest
                Request info for a request. Retrieved from ``ESIMetadata.__getitem__``.
            method: str
                User input request method.
        """
        if method not in self.SUPPORTED_METHODS:
            raise NotImplementedError(f"Request method {method} is not supported yet.")

        if method in metadata.supported_methods:
            return

        logger.warning("Invalid request method: %s for %s", method, metadata.endpoint)
        raise ValueError(
            f"Request method {method} is not supported by endpoint {metadata.endpoint}."
        )



"""Contains all shared OAuth 2.0 flow functions for examples

This module contains all shared functions between the two different OAuth 2.0
flows recommended for web based and mobile/desktop applications. The functions
found here are used by the OAuth 2.0 examples contained in this project.

Source: https://github.com/esi/esi-docs/blob/master/examples/python/sso/shared_flow.py
"""
import requests
import time
import urllib

from .validate_jwt import validate_eve_jwt
from .utils import to_clipboard
from esi_requests.log import getLogger


logger = getLogger(__name__)


def generate_auth_url(client_id, code_challenge=None, **kwd):
    """Generates the URL for users to visit.

    Args:
        client_id: The client ID of an EVE SSO application
        code_challenge: A PKCE code challenge
    """

    redirect = kwd.get("callbackURL", "https://localhost/callback/")
    scope = kwd.get("scope")

    base_auth_url = "https://login.eveonline.com/v2/oauth/authorize/"
    params = {
        "response_type": "code",
        "redirect_uri": redirect,
        "client_id": client_id,
        "scope": scope,
        "state": "unique-state",
    }

    if code_challenge:
        params.update(
            {"code_challenge": code_challenge, "code_challenge_method": "S256"}
        )

    string_params = urllib.parse.urlencode(params)
    full_auth_url = "{}?{}".format(base_auth_url, string_params)

    # copy auth url to clipboard
    to_clipboard(full_auth_url)


def send_token_request(form_values, add_headers={}):
    """Sends a request for an authorization token to the EVE SSO.

    Args:
        form_values: A dict containing the form encoded values that should be
                     sent with the request
        add_headers: A dict containing additional headers to send
    Returns:
        requests.Response: A requests Response object
    """

    headers = {
        "Content-Type": "application/x-www-form-urlencoded",
        "Host": "login.eveonline.com",
    }

    if add_headers:
        headers.update(add_headers)

    res = requests.post(
        "https://login.eveonline.com/v2/oauth/token",
        data=form_values,
        headers=headers,
    )

    res.raise_for_status()
    return res


def handle_sso_token_response(sso_response: requests.Response):
    """Handles the authorization code response from the EVE SSO.

    Args:
        sso_response: A requests Response object gotten by calling the EVE
                      SSO /v2/oauth/token endpoint
    """

    if sso_response.status_code == 200:
        data = sso_response.json()
        access_token = data["access_token"]
        data["retrieve_time"] = int(time.time())

        jwt = validate_eve_jwt(access_token)

        # Find character name
        character_name = jwt["name"]
        data["character_name"] = character_name

        # Find character_id
        sub = jwt["sub"]
        character_id = int(sub.split(":")[-1])
        data["character_id"] = character_id

        return data
    else:
        logger.error("SSO token response error.")
        logger.error("Sent request with url: %s", sso_response.request.url)
        logger.error("Sent request with body: %s", sso_response.request.body)
        logger.error("Sent request with headers: %s", sso_response.request.headers)
        logger.error("SSO response code is: %s", sso_response.status_code)
        logger.error("SSO response JSON is: %s", sso_response.json())
        sso_response.raise_for_status()



import os
import json
import time
from dataclasses import dataclass, asdict
from typing import Optional, List

from .application import Application
from .refresh_token import refresh_token
from .esi_oauth_native import esi_oauth_local
from esi_requests.log import getLogger


logger = getLogger(__name__)


@dataclass
class Token:
    """Hold info for a single token.

    ``(clientId, character_name)`` together act as primary key for each Token.
    """

    access_token: str
    retrieve_time: str
    refresh_token: str
    character_name: str
    character_id: int
    clientId: str


class ESITokens(object):
    """Holds tokens for an esi ``Application``.

    Each Application has a unique ``clientId``, and each Token has a unique ``(clientId, character_name)`` pair.
    The ``ESITokens`` class hold all ``Token``(s) for an ``Application`` (same ``clientId``), 
    so each ``Token`` can be accessed using ``character_name``.

    Customized ``Application`` is supported,
    but the caller needs to make sure the attribute "Application.app" is present in local ``application.json`` file 
    (guaranteed by ``ESI.add_app_generate_token()`` method).

    Attributes:
        app: ``Application``
            An instance of ``Application`` that holds an effective ``clientId``, retrieved from ESI Developer site.
            All tokens generated will have this ``clientId``, along with ``character_name``, as the primary key.

    Keywords:
        update_time: int (seconds)
            Refresh ``Token``(s) if at least update_time passed since last refresh.
            Default 1198 seconds. EVE ESI tokens are effective for 1199 seconds after last refresh.

    Example usage:
    >>> app = Application(clientId, scope, callbackURL)
    >>> with ESITokens(app) as tokens:
    >>>     # do something, such as generate token:
    >>>     tokens.generate()]

    >>> # or use the class without with:
    >>> tokens = ESITokens(app)
    >>> # do something
    >>> tokens.save()   # store change to local file

    Note:
        This class effectively performs one read when initializing, and one read & write upon exiting.
        Calling ``refresh`` and ``generate`` first change ``self.tokens`` buffer, 
        and local ``token.json`` file is updated when calling ``save()`` or use ``with`` statement.
    """

    def __init__(self, app: Application, **kwd):
        self.app = app
        self.clientId = app.clientId
        self.scope = app.scope
        self.callbackURL = app.callbackURL  # not implemented

        self.tokens: List["Token"] = []  # list of tokens for the App

        # ESI token lifespane is 1199 seconds, 1198 for safety
        self._update_time = kwd.get("update_time", 1198)

        # Flag for whether to call save() or not at exit.
        # Every method that changes self.tokens set this to True.
        self._save_flag = False

        self.token_path = os.path.join(os.path.dirname(__file__), "token.json")

        self.__load_tokens()

    def __getitem__(self, cname: str) -> Token:
        """Gets an instance of ``Token`` with ``character_name`` = ``cname``.

        Caller can pass in ``cname = "any"`` to indicate getting any Token for an application.
        Token is refreshed using ``ESITokens.refresh`` before return if the "update_time" threshold is met.
        If current token is invalid, ``ESITokens.generate`` is used to generate a new one.

        Args:
            cname: A string of the character name, acting as a key for a Token.

        Returns:
            A ``Token`` with cname from Application.

        Raises:
            ValueError: No Token matches character_name = {cname}.
        """
        if cname == "any" and self.tokens:
            token = self.tokens[0]
            if self.refresh(token.character_name) is False:
                token = self.generate()
            return token
        for token_ in self.tokens:
            if token_.character_name == cname:
                if self.refresh(token_.character_name) is False:
                    return self.generate()
                return token_

        if cname == "any":
            raise ValueError(f"No Token found.")
        else:
            raise ValueError(f"No Token matches character_name = {cname}.")

    def __str__(self) -> str:
        return "Tokens(app={app}, tokens={tokens})".format(app=str(self.app), tokens=str(self.tokens))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.save()

    def refresh(self, cname: Optional[str] = None) -> bool:
        """Refreshes oauth token of a character for an Application.

        Updates ``access_token`` field of the ``Token`` instance with ``character_name = cname`` stored inside Tokens.
        If parameter ``cname`` is None, refresh all ``Token`` under the Application.
        This method restricts at least 20 minutes between consecutive refresh operations of the same token.

        Args:
            cname: A string of the character name, acting as a key for a Token.

        Returns:
            Whether ALL tokens with ``cname`` have been refreshed successfully.

        Raises:
            KeyError: {cname} is not a valid character name.

        """
        if cname:
            tokens_unrefreshed = [token_ for token_ in self.tokens if token_.character_name == cname]
        else:
            tokens_unrefreshed = self.tokens

        if not tokens_unrefreshed:
            raise KeyError(f"Can't find Token with character_name = {cname}")

        for token_ in tokens_unrefreshed:
            if (int(time.time()) - token_.retrieve_time) < self._update_time:
                # Only refresh a token if 20 minutes have elasped since last refresh
                continue
            new_token_dict = refresh_token(token_.refresh_token, self.clientId)
            if self.__check_refresh_token(new_token_dict) is False:
                # If the Oauth ``refresh_token`` procedure gives an error
                return False

            token_.access_token = new_token_dict["access_token"]
            token_.retrieve_time = new_token_dict["retrieve_time"]
            token_.refresh_token = new_token_dict["refresh_token"]
            # character_name and clientId field should not change.
            self._save_flag = True
        logger.debug("Refresh token successful")
        return True

    def generate(self) -> Token:
        """Generates new token for the Application.

        Generates a new ``Token`` instance that could be used for authenticated request.
        If the token is generated with the same character as an existing ``Token`` for the Application,
        the old ``Token`` in self.tokens will be updated without creating a new Token.

        A url will be copied to clipboard after calling, and the user needs to manually
        visit the url in any browser, complete the EVE login process,
        and copy the URL after login to the command prompt.

        Note:
            New ``Token`` is stored in a buffer, not immediately stored to the file system.

        Args:
            print_info: A bool of whether to print intermediate information in the authorization.
        """
        new_token_dict = esi_oauth_local(
            clientID=self.clientId,
            scope=self.scope,
            callbackURL=self.callbackURL,
        )

        # When user repeatedly calling generate()...
        old_token = None
        for token_ in self.tokens:
            if token_.character_id == new_token_dict["character_id"]:
                old_token = token_
                break

        self._save_flag = True

        if old_token:
            old_token.access_token = new_token_dict["access_token"]
            old_token.retrieve_time = new_token_dict["retrieve_time"]
            old_token.refresh_token = new_token_dict["refresh_token"]
            ret = old_token
        else:
            new_token = Token(
                new_token_dict["access_token"],
                new_token_dict["retrieve_time"],
                new_token_dict["refresh_token"],
                new_token_dict["character_name"],
                new_token_dict["character_id"],
                self.clientId,
            )
            self.tokens.append(new_token)
            ret = new_token

        logger.debug("Generate token successful")
        return ret

    def save(self) -> None:
        """Saves tokens to a local file.

        Packs each ``Token`` in ``self.tokens`` to a dict, and store to local file using json.
        Local file has ``clientId: List[dict, dict, ...]`` pattern.
        """
        if not self.tokens:
            return

        if not self._save_flag:
            return

        tokens_list = [asdict(token_) for token_ in self.tokens]

        if os.path.exists(self.token_path) and os.stat(self.token_path).st_size:
            with open(self.token_path, "r") as all_tokens_fp:
                all_tokens = json.load(all_tokens_fp)
            all_tokens.update({self.clientId: tokens_list})
        else:
            all_tokens = {self.clientId: tokens_list}

        with open(self.token_path, "w") as all_tokens_fp:
            json.dump(all_tokens, all_tokens_fp)
        logger.debug("Save ESITokens successful")

    def exist(self, cname: Optional[str] = None) -> bool:
        """Checks if a ``Token`` or tokens exist or not.

        If ``cname`` is not given, check if current ``Application`` has any token.
        If ``cname`` is given, check if current ``Application`` has a Token with ``character_name = cname``.

        Args:
            cname: A string of the character name, acting as a key for a Token.

        Returns:
            A bool showing if a ``Token`` with ``character_name = cname`` exists or not.
        """
        if cname == "any" or not cname:
            return bool(self.tokens)
        else:
            for token_ in self.tokens:
                if token_.character_name == cname:
                    return True
        return False

    def remove(self, cname: str) -> Token:
        """Removes a Token with given character name.

        Removes a Token with character_name = cname. The Token is removed using list.pop() and returned.
        If no Token matches the cname, raise ValueError.

        Args:
            cname: A string of the character name, acting as a key for a Token.

        Returns:
            Token removed from Application.

        Raises:
            ValueError: No Token matches character_name = {cname}.
        """
        for i in range(len(self.tokens)):
            if self.tokens[i].character_name == cname:
                self._save_flag = True
                return self.tokens.pop(i)

        raise ValueError(f"No Token matches character_name = {cname}.")

    def __load_tokens(self) -> None:
        """Load tokens from local file.

        Read a local json file and search for file[clientId] = self.clientId.
        Unpack tokens in local file to multiple Token(s) stored in self.tokens.
        Does not support custom token file yet. Called once upon init.
        """
        if not os.path.exists(self.token_path) or not os.stat(self.token_path).st_size:
            return

        with open(self.token_path, "r") as all_tokens_fp:
            all_tokens = json.load(all_tokens_fp)

        if self.clientId not in all_tokens:
            return

        tokens = all_tokens[self.clientId]
        for token_ in tokens:
            # More robust when ADDING (removing not considered) new attr to Token but local token.json not updated.
            _append = True
            for attr in Token.__annotations__:
                if attr not in token_:
                    _append = False
                    break

            if _append:
                self.tokens.append(Token(**token_))  # dictionary unpacking

    @staticmethod
    def __check_refresh_token(new_refresh_token: dict) -> bool:
        return "error" not in new_refresh_token and (
            "access_token" in new_refresh_token
            and "retrieve_time" in new_refresh_token
            and "refresh_token" in new_refresh_token
        )



"""
esi_requests.api
~~~~~~~~~~~~~~~~

This module provides simple interface for user.

(c) 2023 by Hanbo Guo
"""
import asyncio
import sys
from typing import TYPE_CHECKING, List, Union

from .sessions import Session

if TYPE_CHECKING:
    from .models import ESIResponse


async def arequest(method: str, endpoint: str, **kwargs):
    """Async version of `request`"""
    async with Session() as session:
        return await session.request(method, endpoint, **kwargs)


def request(method: str, endpoint: str, **kwargs) -> Union["ESIResponse", List["ESIResponse"]]:
    """Make a request to ESI API. 
    
    Uses asyncio.run to run the async function `arequest`.
    """
    # To get away with the error: RuntimeError('Event loop is closed')
    if sys.platform == "win32":
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

    r = asyncio.run(arequest(method, endpoint, **kwargs))
    return r


def get(endpoint: str, params=None, **kwargs):
    """Make a GET request to ESI API."""
    return request("get", endpoint, params=params, **kwargs)


def head(endpoint: str, params=None, **kwargs):
    """Make a HEAD request to ESI API."""
    return request("head", endpoint, params=params, **kwargs)



"""
Test cases for ESIRequestChecker and ESIEndpointChecker.

Contributed by ChatGPT.

(c) 2023 by Hanbo Guo
"""

import os
import json
import unittest
from unittest.mock import patch, Mock
from esi_requests.checker import ESIEndpointChecker


class TestESIEndpointChecker(unittest.TestCase):
    """Checks the status of an ESI endpoint.

    Example:
        ```python
        checker = ESIEndpointChecker()
        if checker('/characters/{character_id}/'):
            print("Endpoint is operational")
        else:
            print("Endpoint is down")
        ```
    """
    def setUp(self):
        self.checker = ESIEndpointChecker()
    
    def tearDown(self):
        del self.checker

    def test_init(self):
        self.assertTrue(self.checker.enabled)
        self.assertTrue(self.checker.fd.name.endswith('status.json'))

    def test_fd_expired(self):
        checker = ESIEndpointChecker()
        checker.status_parsed = None
        self.assertTrue(checker.fd_expired)
        checker.status_parsed = {'fake_endpoint': True}
        self.assertFalse(checker.fd_expired)

    @patch('esi_requests.checker.requests.get')
    def test___call__(self, mock_get):
        # set up the mock response
        route = '/v1/characters/123/'
        mock_status = [{'route': route, 'status': 'green'}]
        mock_response = Mock()
        mock_response.json.return_value = mock_status
        mock_get.return_value = mock_response

        # test with no existing status.json file
        checker = ESIEndpointChecker()
        checker.fd_path = os.path.join(os.path.dirname(__file__), "status.json")
        if os.path.isfile(checker.fd_path):
            os.remove(checker.fd_path)
        self.assertTrue(checker(route))
        self.assertEqual(checker.status_parsed, {route: True})
        self.assertTrue(os.path.isfile(checker.fd_path))

        # test with existing status.json
        checker = ESIEndpointChecker()
        with open(checker.fd_path, "w") as f:
            json.dump({route: False}, f)
        self.assertFalse(checker(route))
        self.assertEqual(checker.status_parsed, {route: True})

    def test_parse_status_json(self):
        status = [{"route": "/v1/characters/123/", "status": "green"}]
        expected_parsed_status = {"/v1/characters/123/": True}
        checker = ESIEndpointChecker()
        parsed_status = checker._parse_status_json(status)
        self.assertEqual(parsed_status, expected_parsed_status)

if __name__ == '__main__':
    unittest.main()




"""Implementation ideas referenced from ESIPy under BSD-3-Clause License"""
import atexit
import inspect
import pickle
from email.utils import parsedate
from datetime import datetime, timedelta
from typing import Union

from .utils import _CacheRecordBaseClass, _CacheRecord, InsertBuffer, _DeleteHandler, hash_key
from .db import ESIDBManager
from esi_requests.log import getLogger

logger = getLogger(__name__)


class BaseCache(_CacheRecordBaseClass):
    """Specifies BaseCache object used by other caching implimentation.

    Requires set and get method for api(s) to access the cache.
    evict is useful for testing. Expired entries should be deleted in get or set,
    so user would not be required to call evict all the time.
    User of BaseCache should call super().__init__() to register instance under cache stats.
    """

    def __init__(self, esidb: ESIDBManager, table: str):
        self.__class__.instances.add(self)
        module = inspect.getmodule(inspect.stack(0)[2][0])
        if module is not None:
            module = module.__name__
        self._record = _CacheRecord(esidb.db_name, table, module)

    def set(self, key, value, expires):
        raise NotImplementedError

    def get(self, key, default):
        raise NotImplementedError

    def evict(self, key):
        raise NotImplementedError

    @property
    def record(self):
        return self._record

    @property
    def hits(self):
        return self._record.hits

    @property
    def miss(self):
        return self._record.miss

    @hits.setter
    def hits(self, val):
        self._record.hits = val

    @miss.setter
    def miss(self, val):
        self._record.miss = val


class SqliteCache(BaseCache):
    """A Sqlite implementation of cache.

    This cache has a key-value interface.
    Specifically, all table has:
    1. ``key``: any ``pickle``able object
    2. ``value``: an arbitrary object serialized by ``pickle.dumps``
    3. ``expires``: retrieved from ``Expires`` headers of HTTP response

    This cache provides a get/set interface.
    Each ``cache.set`` evicts expired cache lines.

    Note:
        This cache does not delete cache line immediately when ``expires`` times is reached.
        It deletes within 5 minutes after this ``expires`` time is reached.
    """

    def __init__(self, name: str, table: str):
        self.cache_name = name
        self.table = table

        self.db = ESIDBManager(name)
        self.buffer = InsertBuffer(self.db)
        self.deleter = _DeleteHandler(self.db, self.table)

        self.db.prepare_table(table, schema="cache")
        atexit.register(self.buffer.flush)
        atexit.register(self.deleter.save)

        self._last_used = None  # used for testing
        super().__init__(self.db, table)

        logger.info("SqliteCache initiated at database %s and table %s", self.db.db_name, table)

    def set(self, key, value, expires: Union[str, int] = 1200):
        """Sets k/v cache line.

        Args:
            key: any ``pickle``able object, e.g. "abcdefg"
            value: any ``pickle``able object
            expires: ``str`` or ``int``. Default expire in 1200 seconds.
                If a ``str`` is provided, it should be in RFC7231 format.
                E.g. "Tue, 11 Jan 2022 10:17:57 GMT"

                If an ``int`` is provided, it should be a number in seconds, e.g. 1200 for 20 minutes.                
        """
        if isinstance(expires, int):
            expires = datetime.utcnow().replace(microsecond=0) + timedelta(seconds=expires)
        else:
            expires = datetime(*parsedate(expires)[:6])

        _h = hash_key(key)
        entry = (_h, pickle.dumps(value), expires)
        self.buffer.insert(entry, self.table)
        self.deleter.update(expires)
        logger.debug("Cache entry set: %s", _h)

    def get(self, key, default=None):
        """Gets the value from cache.

        Attempts to return value with key from cache. A default value is returned if unsuccessful.
        If the selected cache line is expired, delete this cache line.

        Args:
            key: any Python pickle(able) object, e.g. string
            default: If cache returns nothing, returns a default value. Default None.
        """
        _h = hash_key(key)
        
        # First select from buffer, then from database
        row = self.buffer.select(_h)  # hash should be unique, so no need use table parameter
        if not row:
            # should use fetchall and select the newest
            row = self.db.execute(f"SELECT * FROM {self.table} WHERE key=?", (_h,)).fetchone()
        if not row:
            logger.debug("Cache MISS: %s", _h)
            self.miss += 1
            return default

        expires = row[2]
        if isinstance(expires, str):  # expires selected from DB is str, selected from buffer is datatime
            expires = datetime.strptime(expires, "%Y-%m-%d %H:%M:%S")

        if datetime.utcnow() > expires:
            logger.debug("Cache EXPIRED: %s", _h)
            self.miss += 1
            self.db.execute(f"DELETE FROM {self.table} WHERE key=?", (_h,))
            self.db.commit()
            return default  # expired
        else:
            self._last_used = _h
            logger.debug("Cache HIT: %s", _h)
            self.hits += 1
            return pickle.loads(row[1])  # value

    def evict(self, key):
        """Deletes cache entry with key. Useful in testing."""
        _h = hash_key(key)
        self.db.execute(f"DELETE FROM {self.table} WHERE key=?", (_h,))
        self.db.commit()
        logger.debug("Cache entry evicted: %s", _h)



"""
esi_requests
~~~~~~~~~~~~

This module provides a simple interface for making HTTP requests to the EVE Online ESI API.

By default, the `Session` class will automatically handle retries for certain error codes, and raise an exception for other error codes. You can customize this behavior by passing a custom error handling function to the `Session` constructor.

This module is not affiliated with or endorsed by CCP Games or EVE Online.

Some docstrings generated by ChatGPT, a large language model trained by OpenAI.

(c) 2023 by Hanbo Guo
"""

import asyncio
from typing import List, Union

import aiohttp
from tqdm.asyncio import tqdm_asyncio

from .checker import ESIRequestChecker, FakeResponseGenerator, ESIEndpointChecker
from .data.etag_cache import ETagCache
from .exceptions import EndpointDownError
from .models import ESIRequest, ESIResponse, PreparedESIRequest
from .parser import ESIRequestParser


class Session:
    """
    A session class for making requests to ESI API using asyncio.

    Usage:
    ```python
    async with Session() as session:
        response = await session.request("get", "/markets/{region_id}/orders/", region_id=10000002)
        # Use ESIResponse object
    ```

    The session automatically handles request validation, ETag caching, and retrying failed requests.

    Args:
        None.

    Attributes:
        None.
    """

    def __init__(self) -> None:
        self.__request_checker = ESIRequestChecker()
        self.__request_parser = ESIRequestParser()
        self.__etag_cache = ETagCache()
        self.__request_generator = FakeResponseGenerator()
        self.__endpoint_checker = ESIEndpointChecker()

        self.__async_session = None
        # self.__async_event_loop = asyncio.get_event_loop()

        # Due to the complex nature of parameter checking, we disable it by default.
        self.__request_checker.enabled = False

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args):
        await self.close()

    async def request(
        self, method: str, endpoint: str, **kwargs
    ) -> Union["ESIResponse", List["ESIResponse"]]:
        """Sends a request to the ESI API and returns one or more ``ESIResponse`` objects.

        Args:
            method (str): The HTTP method to use (e.g. "GET", "POST").
            endpoint (str): The endpoint to send the request to (e.g. "/characters/{character_id}/assets/").
            **kwargs: Optional arguments to pass to the request, such as query parameters.
        
        Keyword Args:
            enable_tqdm (bool): Whether to enable tqdm progress bar. Defaults to True.
            enable_etag_cache (bool): Whether to enable ETag cache. Defaults to True.
            max_concurrent (int): The maximum number of concurrent requests to send. Defaults to 50.

        Returns:
            Either a single ``ESIResponse`` object, or a list of ``ESIResponse`` objects if multiple requests were issued.

        Note:
            This method automatically handles request validation, ETag caching, and retrying failed requests.
        """
        headers = kwargs.pop("headers", {})
        params = kwargs.pop("params", {})

        # Checks if the endpoint is valid
        if not self.__endpoint_checker.check(endpoint):
            raise EndpointDownError(endpoint)

        request = ESIRequest(endpoint=endpoint, method=method, params=params, headers=headers, **kwargs)

        preps = self.prepare_request(request)

        result = await self.issue(preps, **kwargs)
        if len(result) == 1:
            return result[0]
        return result

    def prepare_request(self, request: "ESIRequest") -> List["PreparedESIRequest"]:
        """Prepares an ``ESIRequest`` for sending to the ESI API.

        Returns:
            A list of ``PreparedESIRequest`` objects, which can be passed to ``Session.issue()``.
        """
        requests = self.__request_parser(request)

        # Update ETag headers
        # see https://developers.eveonline.com/blog/article/esi-etag-best-practices.
        for req in requests:
            if "If-None-Match" not in req.headers:
                req.headers["If-None-Match"] = self.__etag_cache.get(req.url, "")

        request.prepared = True
        return requests

    async def issue(self, requests: List["PreparedESIRequest"], **kwargs) -> List["ESIResponse"]:
        """Issues one or more prepared requests to the ESI API.

        Returns:
            List[ESIResponse]: A list of ``ESIResponse`` objects corresponding to each prepared request.

        Raises:
            ValueError: If a request in the provided list is not an instance of PreparedESIRequest.
        """
        if not isinstance(requests[0], PreparedESIRequest):
            raise ValueError(
                "You can only issue PreparedESIRequest. Use Session.prepare_request() before issue."
            )

        if self.__async_session is None:
            self.__async_session = aiohttp.ClientSession(connector=aiohttp.TCPConnector(ssl=False))

        tasks = [asyncio.create_task(self.issue_one_request(request)) for request in requests]

        # Partition tasks into chunks of size max_concurrent
        max_concurrent = kwargs.get("max_concurrent", 50)
        tasks_partition = [tasks[i : i + max_concurrent] for i in range(0, len(tasks), max_concurrent)]

        result = []

        for tasks_chunk in tasks_partition:
            if kwargs.get("enable_tqdm", True):
                await tqdm_asyncio.gather(*tasks_chunk)  # print progress bar
            else:
                await asyncio.gather(*tasks_chunk)
            result.extend([task.result() for task in tasks_chunk])

        return result

    async def issue_one_request(self, request: "PreparedESIRequest") -> "ESIResponse":
        """Issues a single prepared request to the ESI API and returns the corresponding ``ESIResponse``.

        Args:
            request (``PreparedESIRequest``): The prepared request to send to the ESI API.

        Returns:
            ESIResponse: The ``ESIResponse`` corresponding to the request.

        This asynchronous method first checks if the request would cause an error on the ESI side.
        If the request would cause an error, an ``ESIResponse`` object will be generated as if the request is sent to ESI.
        Otherwise, the method will send the request, and cache the ``ESIResponse`` if necessary.
        """
        # Checks if request would cause error on ESI side
        # If checker is disabled, it will always return True
        valid = await self.__request_checker(request)

        if not valid:
            # If the request would cause error, the request would not be sent to ESI.
            # But still, a ESIResponse will be generated, and this request "appears" to be sent to ESI.
            resp: "ESIResponse" = self.__request_generator.generate(request)

        else:
            # First send the request, then check for ETag
            r = await self.send(request)

            if r.status == 304:
                cache_line = self.__etag_cache.get(request.url)

                if cache_line is None:
                    # TODO: I don't know how to handle cache miss when 304,
                    # if someone could help it would be appreciated
                    raise NotImplementedError(
                        "Not sure how to handle this situation... Please create a Issue on GitHub."
                    )

                resp = cache_line.response

            else:
                resp = await self.build_response(request, r)
                if resp.ok:
                    self.cache_response(resp)

        return resp

    async def send(self, request: "PreparedESIRequest", **kwargs):
        """Sends a prepared request to the ESI API."""
        if isinstance(request, ESIRequest):
            raise ValueError("You can only send ESIRequest that is prepared by Session.prepare.")

        async with self.__async_session.request(
            request.method, request.url, headers=request.headers
        ) as resp:
            await resp.read()
            return resp

    async def build_response(self, request: "ESIRequest", r: "aiohttp.ClientResponse") -> "ESIResponse":
        """Builds an ``ESIResponse`` object from an ``aiohttp.ClientResponse`` object."""
        resp = ESIResponse()

        # aiohttp always gives a status
        resp.status = r.status

        resp.headers = dict(r.headers)  # ETag and Expires live in this headers
        resp.reason = r.reason
        resp.url = r.url
        resp.request_info = request

        # when the response content expires, specified by ESI
        resp.expires = resp.headers.get("Expires")

        resp.text = await r.text()

        return resp

    def cache_response(self, resp: "ESIResponse"):
        """Caches an ``ESIResponse`` object in the ETag cache."""
        etag = resp.headers.get("ETag", "*")
        expires = resp.headers.get("Expires", 24 * 3600)  # 1 day
        self.__etag_cache.set(resp.url, etag=etag, response=resp, expires=expires)

    async def close(self):
        if self.__async_session is None:
            return

        await self.__async_session.close()
        await self.__request_checker.close()

        # if not self.__async_session.closed:
        #     if self.__async_session._connector_owner:
        #         self.__async_session._connector._close()  # silence deprecation warning
        #     self.__async_session._connector = None

        # if not self.__async_event_loop.is_closed():
        #     self.__async_event_loop.run_until_complete(asyncio.sleep(0))
        #     self.__async_event_loop.close()



import pyperclip as pc
import sys
from subprocess import check_call, CalledProcessError, DEVNULL

from esi_requests.log import getLogger

logger = getLogger(__name__)


def to_clipboard(msg: str) -> None:
    """Copies msg to clipboard.

    Copies msg to clipboard using Pyperclip package.
    By default, copy should work on Windows and MacOS.
    Linux needs one of xclip/xsel/gtk/PyQt4 to make it work.
    Some Linux distributions might not have any of these,
    so also tries to install xclip or xsel if possilbe.
    """
    if sys.platform == "linux":  # check xclip/xsel
        xclip_installed = debian_package_check("xclip")
        xsel_installed = debian_package_check("xsel")
        dependency_satisfied = xclip_installed or xsel_installed
        if not xclip_installed and not dependency_satisfied:
            dependency_satisfied = debian_package_install("xclip")
        if not xsel_installed and not dependency_satisfied:
            dependency_satisfied = debian_package_install("xsel")

    try:
        pc.copy(msg)
        logger.debug("Copy msg to clipboard successful: %s", msg)
    except pc.PyperclipException as pc_exc:
        if sys.platform == "linux":  # linux2 prior to Python 3.3
            if not dependency_satisfied:
                logger.error(
                    "Pyperclip NotImplementedError: needs copy/paste mechanism for Linux: xclip or xsel"
                )
                raise SystemExit(
                    "With linux, one of xclip, xsel, gtk, PyQt4 is necessary. apt-get install xclip and xsel failed. Try to manually install them using sudo apt-get install, or get gtk or PyQt4 modules installed. See https://pypi.org/project/pyperclip/ for more."
                ) from pc_exc
        # pc.copy() should work in MacOS and Windows by default
        raise


def debian_package_check(name: str) -> bool:
    """Checks if a debian package is installed.
    Should only be called under Linux system."""
    if sys.platform != "linux":
        raise NotImplemented
    try:
        import apt
    except ImportError as exc:
        if not debian_package_install("python3-apt") and not debian_package_install("python-apt"):
            raise SystemExit("Missing package apt. Install using sudo apt-get install python3-apt or sudo apt-get install python-apt.") from exc
        else:
            # python3-apt or python-apt installed
            import apt
    db_packages = apt.Cache()
    package = db_packages.get(name)
    return package is not None and package.is_installed


def debian_package_install(name: str) -> bool:
    """Tries to install a debian package using sudo apt-get install.

    For some users, this could be successful because they run as sudo.
    In case users does not grant sudo to Python, a password prompt might appear,
    or the call might fail, depending on the Python interpreter.

    Note:
        Run check_call with IDLE will fail. Run with command line, or in VS Code.
    """
    try:
        cmd = "sudo apt-get install -y {}".format(name)
        check_call(
            cmd.split(" "),
            stdout=DEVNULL,
            stderr=DEVNULL,
        )
        logger.info("Installed xclip using: %s", cmd)
        return True
    except CalledProcessError as grepexc:
        logger.error(
            f"FAILED Package install: {grepexc.cmd}: {grepexc.returncode} - {grepexc.output}"
        )
        return False


def read_clipboard() -> str:
    """Reads clipboard using Pyperclip."""
    return pc.paste()



import json
import os
from dataclasses import asdict, dataclass

from esi_requests.log import getLogger

logger = getLogger(__name__)


@dataclass
class Application:
    """Hold info for an Application.

    clientId is the primary key for each Application.
    Application has a save() function, which is useful for storing single Application.
    If multiple Application(s) need to be stored, append to ESIApplications and use its save().

    Attributes:
        clientId: str
            A str acting as a unique key for an Application, created and retrieved from ESI developer site:
            https://developers.eveonline.com/

        scope: str
            A comma seperated str, usually copied from ESI developer site,
            in the format "esi...v1 esi...v1 esi...v1"
        callbackURL: str
            A str of URL that the authentication will redirect to.
            Default and recommend using: https://localhost/callback/

    Example usage:
        >>> # 1. Navigate and log in to https://developers.eveonline.com/.
        >>> # 2. Click "Manage Applications" in the center.
        >>> # 3. Click "Create a new Application" or "View Application".
        >>> # 4. After creating an application or viewing an application,
        >>> #   copy the "Client ID", "Callback URL", and click "Copy Scopes to Clipboard".
        >>> #   Do not reveal "Secret Key" field.
        >>> clientId = {"Client ID" field copied}
        >>> scope = {Pasted content after clicking "Copy Scopes to Clipboard"}
        >>> callbackURL = {"Callback URL" field copied}
        >>> app = Application(clientId, scope, callbackURL)
        >>> # Now you can save() or append to ESIApplications
        >>> app.save()  # save to local file immediately
    """

    clientId: str
    scope: str
    callbackURL: str = "https://localhost/callback/"

    def save(self) -> None:
        """Save current Application to local file.

        If local file already has an Application with the same clientId,
        scope and callbackURL fields will be updated.

        One read and one write per call. Recommend using ESIApplications.save()
        """
        if os.path.exists("application.json") and os.stat("application.json").st_size:
            with open("application.json", "r") as all_apps_fp:
                all_apps = json.load(all_apps_fp)

            old_app = None
            for app_ in all_apps:
                if app_["clientId"] == self.clientId:
                    old_app = app_

            if old_app:
                old_app["callbackURL"] = self.callbackURL
                old_app["scope"] = self.scope
            else:
                all_apps.append(asdict(self))
        else:
            all_apps = [asdict(self)]

        with open("application.json", "w") as all_apps_fp:
            json.dump(all_apps, all_apps_fp)
        logger.info("Save Application successful")


class ESIApplications(object):
    """Hold all Application(s) available.

    Load load application.json file and parse into a list of Application.
    New Application can be appended like list append.
    """

    def __init__(self) -> None:
        self.apps = []

        self.__load_apps()

    def search_scope(self, scope: str) -> Application:
        """Find the first Application with scope.

        Args:
            scope: A str of a single scope. Multiple scope search will yield no Application.

        Returns:
            An Application with the given scope.

        Raises:
            ValueError: No application with scope found.
        """
        for app_ in self.apps:
            scopes = app_.scope.split(" ")
            if scope in scopes:
                return app_

        logger.error("FAILED no application with %s", scope)
        raise ValueError(
            f"No Application with {scope} found. Create one using ESIClient.add_app_generate_token() method."
        )

    def append(self, app: Application) -> None:
        """Append an Application to the ESIApplications instance.

        Args:
            app: An Application. Fields are not checked when append.
        """
        self.apps.append(app)

    def save(self) -> None:
        """Save all Application(s) to a local file.

        Unpack all Application and store them to local application.json file.
        Old file is truncated because the ESIApplications instance is initiated with all Application in the file.
        """
        if not self.apps:
            return

        app_list = [asdict(app_) for app_ in self.apps]
        with open("application.json", "w") as all_apps_fp:
            json.dump(app_list, all_apps_fp)
        logger.info("Save ESIApplications successful")

    def __load_apps(self) -> None:
        if not os.path.exists("application.json") or not os.stat("application.json").st_size:
            return

        with open("application.json", "r") as all_apps_fp:
            all_apps = json.load(all_apps_fp)

        for app_ in all_apps:
            self.apps.append(Application(**app_))  # dict unpacking



import os
import logging
import sys

if sys.platform == "win32":
    from concurrent_log_handler import ConcurrentRotatingFileHandler
from logging import Logger
from logging.handlers import RotatingFileHandler
from typing import Optional


FORMATTER = logging.Formatter(
    "%(asctime)s %(levelname)s %(name)s@%(lineno)d: %(message)s"
)
MAXBYTES = 5 * 1024 * 1024
BACKUPCOUNT = 10

LOGLEVEL = 30           # WARNING
LOGFILE = "esi.log"     # default filename

# Log levels:
# DEBUG: detailed information, clear workflow
# INFO: conceptual information
# WARNING: unexpected or indicative of future problems
# ERROR: failed some functions, but not crashing
# CRITICAL: crashing


def get_stream_handler():
    handler = logging.StreamHandler()  # stderr
    handler.setLevel(logging.CRITICAL)
    handler.setFormatter(FORMATTER)
    return handler


def getLogger(
    name: str, filename: Optional[str] = ..., level: Optional[int] = ...
) -> Logger:
    """Returns a logger with specified name and level, using RotatingFileHandler with filename for streaming.

    Args:
        name: str
            Logger name. Used in logging.getLogger(name).
        filename: str | None
            File stream target for RotatingFileHandler. filename is joined using eve_tools/log/{filename}.
            If not given, default "esi.log".
        level: int | None
            Level of the logger. Default WARNING. Default value configured in eve_tools/config/__init__.
    """
    logger = logging.getLogger(name)

    if level is Ellipsis:
        level = LOGLEVEL
    logger.setLevel(level)

    if filename is Ellipsis:
        filename = LOGFILE
    filename = os.path.realpath(os.path.join(os.path.dirname(__file__), filename))

    if sys.platform == "win32":
        file_handler = ConcurrentRotatingFileHandler(filename, maxBytes=MAXBYTES, backupCount=BACKUPCOUNT, delay=True)
    else:  # RotatingFileHandler works on linux/darwin
        file_handler = RotatingFileHandler(
            filename, maxBytes=MAXBYTES, backupCount=BACKUPCOUNT, delay=True  # 5MB * 10
        )
    file_handler.setLevel(level)
    file_handler.setFormatter(FORMATTER)

    logger.addHandler(file_handler)
    logger.addHandler(get_stream_handler())

    logger.propagate = False  # intuitively not necessary

    return logger
